#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
import numpy as np
import pyomo.environ as pyo
import idaes.logger as idaeslog
from pyomo.util.infeasible import log_infeasible_constraints, log_infeasible_bounds, log_close_to_bounds
from idaes.apps.grid_integration.multiperiod.multiperiod import MultiPeriodModel
from dispatches.case_studies.renewables_case.RE_flowsheet import *


def wind_battery_pem_tank_turb_variable_pairs(m1, m2, tank_type):
    """
    This function links together unit model state variables from one timestep to the next.

    The hydrogen tank and the battery model have material and energy holdups that need to be consistent across time blocks.
    If using the `simple` tank model, there are no energy holdups to account for. For the `detailed` tank model, the emergy
    holdups need to be linked.

    Args:
        m1: current time block model
        m2: next time block model
        tank_type: `simple`, `detailed` or `detailed-valve`
    """
    if tank_type == "simple":
        pairs = [(m1.fs.h2_tank.tank_holdup[0], m2.fs.h2_tank.tank_holdup_previous[0])]
    else:
        pairs = [(m1.fs.h2_tank.material_holdup[0, ('Vap', 'hydrogen')],
                m2.fs.h2_tank.previous_material_holdup[0, ('Vap', 'hydrogen')])]
        if tank_type == 'detailed-valve':
            pairs += [(m1.fs.h2_tank.energy_holdup[0, 'Vap'], m2.fs.h2_tank.previous_energy_holdup[0, 'Vap'])]
    pairs += [(m1.fs.battery.state_of_charge[0], m2.fs.battery.initial_state_of_charge),
              (m1.fs.battery.energy_throughput[0], m2.fs.battery.initial_energy_throughput),
              (m1.fs.battery.nameplate_power, m2.fs.battery.nameplate_power)]
    
    return pairs


def wind_battery_pem_tank_turb_periodic_variable_pairs(m1, m2, tank_type):
    """
    The final hydrogen material holdup and battery storage of charge must be the same as in the intial timestep. 
    If using the `simple` tank model, there are no energy holdups to account for. For the `detailed` tank model, the emergy
    holdups need to be linked.

    Args:
        m1: final time block model
        m2: first time block model
        tank_type: `simple`, `detailed` or `detailed-valve`
    """
    if tank_type == "simple":
        pairs = [(m1.fs.h2_tank.tank_holdup[0], m2.fs.h2_tank.tank_holdup_previous[0])]
    else:
        pairs = [(m1.fs.h2_tank.material_holdup[0, ('Vap', 'hydrogen')],
                m2.fs.h2_tank.previous_material_holdup[0, ('Vap', 'hydrogen')])]
        if tank_type == 'detailed-valve':
            pairs += [(m1.fs.h2_tank.energy_holdup[0, 'Vap'], m2.fs.h2_tank.previous_energy_holdup[0, 'Vap'])]
    pairs += [(m1.fs.battery.state_of_charge[0], m2.fs.battery.initial_state_of_charge),
              (m1.fs.battery.nameplate_power, m2.fs.battery.nameplate_power)]
    return pairs


def wind_battery_pem_tank_turb_om_costs(m):
    """
    Add unit fixed and variable operating costs as parameters for the unit model m
    """
    m.fs.windpower.op_cost = pyo.Param(
        initialize=wind_op_cost,
        doc="fixed cost of operating wind plant $/kW-yr")
    m.fs.pem.op_cost = pyo.Param(
        initialize=pem_op_cost,
        doc="fixed cost of operating pem $/kW-yr"
    )
    m.fs.pem.var_cost = pyo.Param(
        initialize=pem_var_cost,
        doc="variable operating cost of pem $/kWh"
    )
    m.fs.h2_tank.op_cost = Expression(
        expr=tank_op_cost,
        doc="fixed cost of operating tank in $/m^3"
    )
    m.fs.h2_turbine.op_cost = Expression(
        expr=turbine_op_cost,
        doc="fixed cost of operating turbine $/kW-pr"
    )
    m.fs.h2_turbine.var_cost = Expression(
        expr=turbine_var_cost,
        doc="variable operating cost of turbine $/kg"
    )


def initialize_fs(m, tank_type, verbose=False):
    """
    Initializing the flowsheet is done starting with the wind model and propagating the solved initial state to downstream models.

    The splitter is initialized with no flow to the battery or PEM so all electricity flows to the grid, which makes the initialization of all
    unit models downstream of the wind plant independent of its time-varying electricity production. This initialzation function can
    then be repeated for all timesteps within a dynamic analysis.

    Args:
        m: model
        tank_type: `simple`, `detailed` or `detailed-valve`
        verbose:
    """
    outlvl = idaeslog.INFO if verbose else idaeslog.WARNING

    m.fs.windpower.initialize(outlvl=outlvl)

    propagate_state(m.fs.wind_to_splitter)
    m.fs.splitter.battery_elec[0].fix(0)
    m.fs.splitter.pem_elec[0].fix(0)
    m.fs.splitter.initialize()
    m.fs.splitter.battery_elec[0].unfix()
    m.fs.splitter.pem_elec[0].unfix()
    if verbose:
        m.fs.splitter.report(dof=True)

    propagate_state(m.fs.splitter_to_pem)
    propagate_state(m.fs.splitter_to_battery)

    m.fs.battery.elec_in[0].fix()
    m.fs.battery.elec_out[0].fix(value(m.fs.battery.elec_in[0]))
    m.fs.battery.initialize(outlvl=outlvl)
    m.fs.battery.elec_in[0].unfix()
    m.fs.battery.elec_out[0].unfix()
    if verbose:
        m.fs.battery.report(dof=True)

    m.fs.pem.initialize(outlvl=outlvl)
    if verbose:
        m.fs.pem.report(dof=True)

    propagate_state(m.fs.pem_to_tank)

    if tank_type == "simple":
        m.fs.h2_tank.outlet_to_turbine.flow_mol[0].fix(value(m.fs.h2_tank.inlet.flow_mol[0]))
        m.fs.h2_tank.outlet_to_pipeline.flow_mol[0].fix(0)
        m.fs.h2_tank.tank_holdup_previous.fix(0)
        m.fs.h2_tank.initialize(outlvl=outlvl)
        m.fs.h2_tank.outlet_to_turbine.flow_mol[0].unfix()
        m.fs.h2_tank.outlet_to_pipeline.flow_mol[0].unfix()
        m.fs.h2_tank.tank_holdup_previous.unfix()
    else:
        m.fs.h2_tank.outlet.flow_mol[0].fix(value(m.fs.h2_tank.inlet.flow_mol[0]))
        m.fs.h2_tank.initialize(outlvl=outlvl)
        m.fs.h2_tank.outlet.flow_mol[0].unfix()
        if verbose:
            m.fs.h2_tank.report(dof=True)

    if tank_type == "simple":
        propagate_state(m.fs.h2_tank_to_turb)
    else:
        if tank_type == "detailed-valve":
            propagate_state(m.fs.tank_to_valve)
            m.fs.tank_valve.initialize(outlvl=outlvl)
            if verbose:
                m.fs.tank_valve.report(dof=True)

        propagate_state(m.fs.valve_to_h2_splitter)
        m.fs.h2_splitter.split_fraction[0, "sold"].fix(0.5)
        m.fs.h2_splitter.initialize(outlvl=outlvl)
        m.fs.h2_splitter.split_fraction[0, "sold"].unfix()
        if verbose:
            m.fs.h2_splitter.report(dof=True)

        propagate_state(m.fs.h2_splitter_to_turb)

    m.fs.translator.initialize(outlvl=outlvl)
    if verbose:
        m.fs.translator.report(dof=True)

    propagate_state(m.fs.translator_to_mixer)
    m.fs.mixer.air_h2_ratio.deactivate()
    m.fs.mixer.purchased_hydrogen_feed.flow_mol[0].fix(h2_turb_min_flow)
    h2_out = value(m.fs.mixer.hydrogen_feed.flow_mol[0] + m.fs.mixer.purchased_hydrogen_feed.flow_mol[0])
    m.fs.mixer.air_feed.flow_mol[0].fix(h2_out * air_h2_ratio)
    m.fs.mixer.initialize(outlvl=outlvl)
    m.fs.mixer.purchased_hydrogen_feed.flow_mol[0].unfix()
    m.fs.mixer.air_feed.flow_mol[0].unfix()
    m.fs.mixer.air_h2_ratio.activate()
    if verbose:
        m.fs.mixer.report(dof=True)

    propagate_state(m.fs.mixer_to_turbine)

    m.fs.h2_turbine.initialize(outlvl=outlvl)
    if verbose:
        m.fs.h2_turbine.report(dof=True)


def wind_battery_pem_tank_turb_model(wind_resource_config, input_params, verbose):
    """
    Creates an initialized flowsheet model for a single time step with operating, size and cost components
    
    First, the model is created using the input_params and wind_resource_config
    Second, the model is initialized so that it solves and its values are internally consistent
    Third, battery ramp constraints and operating cost components are added

    Args:
        wind_resource_config: wind resource for the time step
        input_params: size and operation parameters. Required keys: `wind_mw`, `pem_bar`, `batt_mw`, `tank_type`, `tank_size`, `pem_bar`
        verbose:
    """
    m = create_model(input_params['wind_mw'], input_params['pem_bar'], input_params['batt_mw'], input_params['tank_type'], input_params['tank_size'], input_params['pem_bar'],
                     wind_resource_config)

    m.fs.battery.initial_state_of_charge.fix(0)
    m.fs.battery.initial_energy_throughput.fix(0)

    if input_params['tank_type'] == "detailed":
        m.fs.h2_tank.previous_state[0].temperature.fix(input_params['pem_temp'])
        m.fs.h2_tank.previous_state[0].pressure.fix(input_params['pem_bar'] * 1e5)

    initialize_fs(m, input_params['tank_type'], verbose=verbose)

    if input_params['tank_type'] == "detailed":
        m.fs.h2_tank.previous_state[0].temperature.unfix()
        m.fs.h2_tank.previous_state[0].pressure.unfix()
    m.fs.battery.initial_state_of_charge.unfix()
    m.fs.battery.initial_energy_throughput.unfix()

    batt = m.fs.battery
    batt.energy_down_ramp = pyo.Constraint(
        expr=batt.initial_state_of_charge - batt.state_of_charge[0] <= battery_ramp_rate)
    batt.energy_up_ramp = pyo.Constraint(
        expr=batt.state_of_charge[0] - batt.initial_state_of_charge <= battery_ramp_rate)

    wind_battery_pem_tank_turb_om_costs(m)

    return m


def wind_battery_pem_tank_turb_mp_block(wind_resource_config, input_params, verbose):
    """
    Wrapper of `wind_battery_pem_tank_turb_model` for creating the process model per time point for the MultiPeriod model.
    Uses cloning of the Pyomo model in order to reduce runtime. 
    The clone is reinitialized with the `wind_resource_config` for the given time point, which only required modifying
    the windpower and the splitter, as the rest of the units have no flow and therefore is unaffected by wind resource changes.

    Args:
        wind_resource_config: dictionary with `resource_speed` for the time step
        input_params: size and operation parameters. Required keys: `wind_mw`, `pem_bar`, `batt_mw`, `tank_type`, `tank_size`, `pem_bar`
        verbose:
    """

    if 'pyo_model' not in input_params.keys():
        input_params['pyo_model'] = wind_battery_pem_tank_turb_model(wind_resource_config, input_params, verbose)
    m = input_params['pyo_model'].clone()

    if 'resource_speed' in wind_resource_config.keys():
        m.fs.windpower.config.resource_speed = wind_resource_config['resource_speed']
    elif 'capacity_factor' in wind_resource_config.keys():
        m.fs.windpower.config.capacity_factor = wind_resource_config['capacity_factor']
    else:
        raise ValueError(f"`wind_resource_config` dict must contain either 'resource_speed' or 'capacity_factor' values")

    m.fs.windpower.setup_resource()

    outlvl = idaeslog.INFO if verbose else idaeslog.WARNING
    m.fs.windpower.initialize(outlvl=outlvl)
    propagate_state(m.fs.wind_to_splitter)
    m.fs.splitter.initialize()
    return m


def wind_battery_pem_tank_turb_optimize(n_time_points, input_params, verbose=False, plot=False):
    """
    The main function for optimizing the flowsheet's design and operating variables for Net Present Value. 

    Creates the MultiPeriodModel and adds the size and operating constraints in addition to the Net Present Value Objective.
    The NPV is a function of the capital costs, the electricity market profit, the hydrogen market profit, and the capital recovery factor.
    The operating decisions and state evolution of the unit models and the flowsheet as a whole form the constraints of the Non-linear Program.

    Required input parameters include:
        `wind_mw`: initial guess of the wind size
        `wind_mw_ub`: upper bound of wind size
        `batt_mw`: initial guess of the battery size
        `pem_mw`: initial guess of the pem size
        `pem_bar`: operating pressure
        `pem_temp`: operating temperature [K]
        `tank_size`: initial guess of the tank_size [kg H2]
        `tank_type`: "simple", "detailed" or "detailed-valve"
        `turb_mw`: intial guess of the turbine size
        `wind_resource`: dictionary of wind resource configs for each time point
        `h2_price_per_kg`: market price of hydrogen
        `DA_LMPs`: LMPs for each time point
        `design_opt`: true to optimize design, else sizes are fixed at initial guess sizes
        `extant_wind`: if true, fix wind size to initial size and do not add wind capital cost to NPV

    Args:
        n_time_points: number of periods in MultiPeriod model
        input_params: 
        verbose: print all logging and outputs from unit models, initialization, solvers, etc
        plot: plot the operating variables time series
    """
    # create the multiperiod model object
    tank_type = input_params['tank_type']
    mp_model = MultiPeriodModel(n_time_points=n_time_points,
                                process_model_func=partial(wind_battery_pem_tank_turb_mp_block, input_params=input_params, verbose=verbose),
                                linking_variable_func=partial(wind_battery_pem_tank_turb_variable_pairs, tank_type=tank_type),
                                periodic_variable_func=partial(wind_battery_pem_tank_turb_periodic_variable_pairs, tank_type=tank_type))

    mp_model.build_multi_period_model(input_params['wind_resource'])

    m = mp_model.pyomo_model
    blks = mp_model.get_active_process_blocks()
    blks[0].fs.battery.initial_energy_throughput.fix(0)
    
    m.wind_system_capacity = Var(domain=NonNegativeReals, initialize=input_params['wind_mw'] * 1e3, units=pyunits.kW, bounds=(0, input_params['wind_mw_ub'] * 1e3))
    m.battery_system_capacity = Var(domain=NonNegativeReals, initialize=input_params['batt_mw'] * 1e3, units=pyunits.kW)
    m.pem_system_capacity = Var(domain=NonNegativeReals, initialize=input_params['pem_mw'] * 1e3, units=pyunits.kW)
    m.h2_tank_size = Var(domain=NonNegativeReals, initialize=input_params['tank_size'])
    m.turb_system_capacity = Var(domain=NonNegativeReals, initialize=input_params['turb_mw'] * 1e3, units=pyunits.kW)

    # add size constraints
    if input_params['design_opt']:
        if input_params['extant_wind']:
            m.wind_system_capacity.fix()
        for blk in blks:
            if tank_type == "detailed":
                blk.fs.h2_tank.tank_length.unfix()
            blk.fs.battery.nameplate_power.unfix()
    else:
        m.pem_system_capacity.fix(input_params['pem_mw'])
        m.h2_tank_size.fix(input_params['tank_size'])
        m.turb_system_capacity.fix(input_params['turb_mw'])

    m.wind_max_p = Constraint(mp_model.pyomo_model.TIME, rule=lambda b, t: blks[t].fs.windpower.system_capacity <= m.wind_system_capacity)
    m.battery_max_p = Constraint(mp_model.pyomo_model.TIME, rule=lambda b, t: blks[t].fs.battery.nameplate_power <= m.battery_system_capacity)
    m.pem_max_p = Constraint(mp_model.pyomo_model.TIME, rule=lambda b, t: blks[t].fs.pem.electricity[0] <= m.pem_system_capacity)
    if tank_type == "simple":
        m.tank_max_p = Constraint(mp_model.pyomo_model.TIME, rule=lambda b, t: blks[t].fs.h2_tank.tank_holdup[0] <= m.h2_tank_size)
    else:
        m.tank_max_p = Constraint(mp_model.pyomo_model.TIME, rule=lambda b, t: blks[t].fs.h2_tank.material_holdup[0, "Vap", "hydrogen"] <= m.h2_tank_size)
    m.turb_max_p = Constraint(mp_model.pyomo_model.TIME, rule=lambda b, t: blks[t].fs.h2_turbine.electricity[0] <= m.turb_system_capacity)

    # Add cost and price equations
    m.wind_cap_cost = pyo.Param(default=wind_cap_cost, mutable=True)
    if input_params['extant_wind']:
        m.wind_cap_cost.set_value(0.)
    m.pem_cap_cost = pyo.Param(default=pem_cap_cost, mutable=True)
    m.batt_cap_cost = pyo.Param(default=batt_cap_cost, mutable=True)
    m.tank_cap_cost = pyo.Param(default=tank_cap_cost_per_kg, mutable=True)
    m.turb_cap_cost = pyo.Param(default=turbine_cap_cost, mutable=True)

    m.contract_capacity = Var(domain=NonNegativeReals, initialize=0, units=pyunits.mol / pyunits.second)
    m.h2_price_per_kg = pyo.Param(default=input_params['h2_price_per_kg'], mutable=True)

    for blk in blks:
        blk_wind = blk.fs.windpower
        blk_battery = blk.fs.battery
        blk_pem = blk.fs.pem
        blk_tank = blk.fs.h2_tank
        blk_turb = blk.fs.h2_turbine

        # calculate operating costs
        blk_wind.op_total_cost = Expression(
            expr=blk_wind.system_capacity * blk_wind.op_cost / 8760,
        )
        blk_pem.op_total_cost = Expression(
            expr=m.pem_system_capacity * blk_pem.op_cost / 8760 + blk_pem.var_cost * blk_pem.electricity[0],
        )
        blk_tank.op_total_cost = Expression(
            expr=m.h2_tank_size * blk_tank.op_cost / 8760
        )
        blk_turb.op_total_cost = Expression(
            expr=m.turb_system_capacity * blk_turb.op_cost / 8760 + blk_turb.var_cost * blk_turb.electricity[0]
        )

        # add market data for each block
        blk.lmp_signal = pyo.Param(default=0, mutable=True)
        blk.revenue = blk.lmp_signal * (blk.fs.splitter.grid_elec[0] + blk_battery.elec_out[0] + blk_turb.electricity[0])
        blk.profit = pyo.Expression(expr=blk.revenue
                                         - blk_wind.op_total_cost
                                         - blk_pem.op_total_cost
                                         - blk_tank.op_total_cost
                                         - blk_turb.op_total_cost
                                    )
        if tank_type == "simple":
            blk.hydrogen_revenue = Expression(expr=m.h2_price_per_kg / h2_mols_per_kg * (
                blk_tank.outlet_to_pipeline.flow_mol[0] - blk.fs.mixer.purchased_hydrogen_feed_state[0].flow_mol) * 3600)
        else:
            blk.hydrogen_revenue = Expression(expr=m.h2_price_per_kg / h2_mols_per_kg * (
                blk.fs.h2_splitter.sold.flow_mol[0] - blk.fs.mixer.purchased_hydrogen_feed_state[0].flow_mol) * 3600)

    for (i, blk) in enumerate(blks):
        blk.lmp_signal.set_value(input_params['DA_LMPs'][i] * 1e-3)     # to $/kWh

    n_weeks = n_time_points / (7 * 24)

    m.annual_revenue = Expression(expr=(sum([blk.profit + blk.hydrogen_revenue for blk in blks])) * 52.143 / n_weeks)

    m.NPV = Expression(expr=-(m.wind_cap_cost * m.wind_system_capacity
                              + m.batt_cap_cost * m.battery_system_capacity
                              + m.pem_cap_cost * m.pem_system_capacity
                              + m.tank_cap_cost * m.h2_tank_size
                              + m.turb_cap_cost * m.turb_system_capacity
                              ) + PA * m.annual_revenue)
    m.obj = pyo.Objective(expr=-m.NPV * 1e-8)


    opt = pyo.SolverFactory('ipopt')

    opt.options['max_iter'] = 10000

    if verbose:
        solve_log = idaeslog.getInitLogger("infeasibility", idaeslog.INFO, tag="properties")
        log_infeasible_constraints(m, logger=solve_log, tol=1e-4, log_expression=True, log_variables=True)
        log_infeasible_bounds(m, logger=solve_log, tol=1e-4)

    opt.solve(m, tee=verbose)

    if verbose:
        solve_log = idaeslog.getInitLogger("infeasibility", idaeslog.INFO, tag="properties")
        log_infeasible_constraints(m, logger=solve_log, tol=1e-4, log_expression=False, log_variables=False)
        log_infeasible_bounds(m, logger=solve_log, tol=1e-4)


    h2_prod = []
    wind_to_grid = []
    wind_to_pem = []
    wind_to_batt = []
    batt_to_grid = []
    soc = []
    wind_gen = []
    h2_tank_in = []
    h2_tank_out = []
    h2_tank_holdup = []
    h2_purchased = []
    h2_turbine_elec = []
    comp_kwh = []
    turb_kwh = []
    h2_revenue = []
    elec_income = []

    h2_prod.append([pyo.value(blks[i].fs.pem.outlet_state[0].flow_mol * 3600 / 500) for i in range(n_time_points)])
    h2_tank_in.append([pyo.value(blks[i].fs.h2_tank.inlet.flow_mol[0] * 3600 / 500) for i in range(n_time_points)])
    if tank_type == "simple":
        h2_tank_out.append([pyo.value((blks[i].fs.h2_tank.outlet_to_pipeline.flow_mol[0] + blks[i].fs.h2_tank.outlet_to_turbine.flow_mol[0]) * 3600 / 500) for i in range(n_time_points)])
        h2_tank_holdup.append([pyo.value(blks[i].fs.h2_tank.tank_holdup[0]) for i in range(n_time_points)])
    else:  
        h2_tank_out.append([pyo.value(blks[i].fs.h2_tank.outlet.flow_mol[0] * 3600 / 500) for i in range(n_time_points)])
        h2_tank_holdup.append(
            [pyo.value(blks[i].fs.h2_tank.material_holdup[0, ('Vap', 'hydrogen')]) for i in range(n_time_points)])
    h2_purchased.append([pyo.value(blks[i].fs.mixer.purchased_hydrogen_feed_state[0].flow_mol) * 3600 / h2_mols_per_kg for i in range(n_time_points)])

    wind_gen.append([pyo.value(blks[i].fs.windpower.electricity[0]) for i in range(n_time_points)])
    wind_to_grid.append([pyo.value(blks[i].fs.splitter.grid_elec[0]) for i in range(n_time_points)])
    wind_to_pem.append([pyo.value(blks[i].fs.pem.electricity[0]) for i in range(n_time_points)])
    batt_to_grid.append([pyo.value(blks[i].fs.battery.elec_out[0]) for i in range(n_time_points)])
    wind_to_batt.append([pyo.value(blks[i].fs.battery.elec_in[0]) for i in range(n_time_points)])
    h2_turbine_elec.append([pyo.value(blks[i].fs.h2_turbine.electricity[0]) for i in range(n_time_points)])
    turb_kwh.append(
        [pyo.value(blks[i].fs.h2_turbine.turbine.work_mechanical[0]) * -1e-3 for i in range(n_time_points)])
    comp_kwh.append(
        [pyo.value(blks[i].fs.h2_turbine.compressor.work_mechanical[0]) * 1e-3 for i in range(n_time_points)])
    elec_income.append([pyo.value(blks[i].profit) for i in range(n_time_points)])
    h2_revenue.append([pyo.value(blks[i].hydrogen_revenue) for i in range(n_time_points)])

    n_weeks_to_plot = 1
    hours = np.arange(n_time_points)
    lmp_array = input_params['DA_LMPs'][0:n_time_points]
    h2_prod = np.asarray(h2_prod[0:n_weeks_to_plot]).flatten()
    wind_to_pem = np.asarray(wind_to_pem[0:n_weeks_to_plot]).flatten()
    wind_gen = np.asarray(wind_gen[0:n_weeks_to_plot]).flatten()
    wind_out = np.asarray(wind_to_grid[0:n_weeks_to_plot]).flatten()
    h2_tank_in = np.asarray(h2_tank_in[0:n_weeks_to_plot]).flatten()
    h2_tank_out = np.asarray(h2_tank_out[0:n_weeks_to_plot]).flatten()
    h2_tank_holdup = np.asarray(h2_tank_holdup[0:n_weeks_to_plot]).flatten()
    batt_out = np.asarray(batt_to_grid[0:n_weeks_to_plot]).flatten()
    batt_in = np.asarray(wind_to_batt[0:n_weeks_to_plot]).flatten()
    h2_purchased = np.asarray(h2_purchased[0:n_weeks_to_plot]).flatten()
    turb_kwh = np.asarray(turb_kwh[0:n_weeks_to_plot]).flatten()
    comp_kwh = np.asarray(comp_kwh[0:n_weeks_to_plot]).flatten()
    h2_turbine_elec = np.asarray(h2_turbine_elec[0:n_weeks_to_plot]).flatten()
    h2_revenue = np.asarray(h2_revenue[0:n_weeks_to_plot]).flatten()
    elec_income = np.asarray(elec_income[0:n_weeks_to_plot]).flatten()

    wind_cap = value(m.wind_system_capacity) * 1e-3
    batt_cap = value(m.battery_system_capacity) * 1e-3
    pem_cap = value(m.pem_system_capacity) * 1e-3
    tank_size = value(m.h2_tank_size)
    turb_cap = value(m.turb_system_capacity) * 1e-3

    design_res = {
        'wind_mw': wind_cap,
        "batt_mw": batt_cap,
        "pem_mw": pem_cap,
        "tank_kgH2": tank_size,
        "turb_mw": turb_cap,
        "avg_turb_eff": np.average(turb_kwh/comp_kwh),
        "annual_rev_h2": sum(h2_revenue) * 52 / n_weeks,
        "annual_rev_E": sum(elec_income) * 52 / n_weeks,
        "NPV": value(m.NPV)
    }

    print(design_res)

    if plot:
        fig, ax1 = plt.subplots(3, 1, figsize=(20, 8))
        fig.suptitle(f"Optimal NPV ${round(value(m.NPV) * 1e-6)}mil from {round(batt_cap, 2)} MW Battery, "
                     f"{round(pem_cap, 2)} MW PEM, {round(tank_size, 2)} kgH2 Tank and {round(turb_cap, 2)} MW Turbine")

        # color = 'tab:green'
        ax1[0].set_xlabel('Hour')
        # ax1[0].set_ylabel('kW', )
        ax1[0].step(hours, wind_gen, label="Wind Generation [kW]")
        ax1[0].step(hours, wind_out, label="Wind to Grid [kW]")
        ax1[0].step(hours, wind_to_pem, label="Wind to Pem [kW]")
        ax1[0].step(hours, batt_in, label="Wind to Batt [kW]")
        ax1[0].step(hours, batt_out, label="Batt to Grid [kW]")
        ax1[0].step(hours, h2_turbine_elec, label="H2 Turbine [kW]")
        ax1[0].tick_params(axis='y', )
        ax1[0].legend()
        ax1[0].grid(visible=True, which='major', color='k', linestyle='--', alpha=0.2)
        ax1[0].minorticks_on()
        ax1[0].grid(visible=True, which='minor', color='k', linestyle='--', alpha=0.2)

        ax2 = ax1[0].twinx()
        color = 'k'
        ax2.set_ylabel('LMP [$/MWh]', color=color)
        ax2.plot(hours, lmp_array[0:len(hours)], color=color)
        ax2.tick_params(axis='y', labelcolor=color)

        # ax1[1].set_xlabel('Hour')
        # ax1[1].set_ylabel('kg/hr', )
        ax1[1].step(hours, h2_prod, label="PEM H2 production [kg/hr]")
        ax1[1].step(hours, h2_tank_in, label="Tank inlet [kg/hr]")
        ax1[1].step(hours, h2_tank_out, label="Tank outlet [kg/hr]")
        ax1[1].step(hours, h2_tank_holdup, label="Tank holdup [kg]")
        ax1[1].step(hours, h2_purchased, label="H2 purchased [kg/hr]")
        ax1[1].tick_params(axis='y', )
        ax1[1].legend()
        ax1[1].grid(visible=True, which='major', color='k', linestyle='--', alpha=0.2)
        ax1[1].minorticks_on()
        ax1[1].grid(visible=True, which='minor', color='k', linestyle='--', alpha=0.2)

        ax2 = ax1[1].twinx()
        color = 'k'
        ax2.set_ylabel('LMP [$/MWh]', color=color)
        ax2.plot(hours, lmp_array[0:len(hours)], color=color)
        ax2.tick_params(axis='y', labelcolor=color)

        ax1[2].set_xlabel('Hour')
        ax1[2].step(hours, elec_income, label="Elec Income")
        ax1[2].step(hours, h2_revenue, label="H2 rev")
        ax1[2].step(hours, np.cumsum(elec_income), label="Elec Income cumulative")
        ax1[2].step(hours, np.cumsum(h2_revenue), label="H2 rev cumulative")
        ax1[2].legend()
        ax1[2].grid(visible=True, which='major', color='k', linestyle='--', alpha=0.2)
        ax1[2].minorticks_on()
        ax1[2].grid(visible=True, which='minor', color='k', linestyle='--', alpha=0.2)
        fig.tight_layout()

    plt.show()

    return design_res


if __name__ == "__main__":
    default_input_params['wind_mw'] = 200
    des_res = wind_battery_pem_tank_turb_optimize(n_time_points=6 * 24, input_params=default_input_params, verbose=False, plot=True)
    print(des_res)


#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
import sys
import pandas as pd
import textwrap
# Import Pyomo libraries
from pyomo.environ import Var, Reals, value, units as pyunits
from pyomo.network import Port
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.util.calc_var_value import calculate_variable_from_constraint

# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        UnitModelBlockData,
                        useDefault)
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.tables import stream_table_dataframe_to_string, create_stream_table_dataframe
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_activated_constraints,
                                              number_activated_blocks)
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)


def _make_pem_electrolyzer_config_block(config):
    config.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""PEM Electrolyzer does not support dynamic models, thus this must be
False."""))
    config.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag",
        doc="""Gibbs reactors do not have defined volume, thus this must be
    False."""))
    config.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}"""))
    config.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))


@declare_process_block_class("PEM_Electrolyzer", doc="Simple 0D proton-exchange membrane electrolyzer model.")
class PEMElectrolyzerData(UnitModelBlockData):
    """
    Simple 0D proton-exchange membrane electrolyzer model.
    Unit model to convert electricity and water into H2 gas.
    """
    CONFIG = ConfigBlock()
    _make_pem_electrolyzer_config_block(CONFIG)

    def build(self):
        """Building model

        Args:
            None
        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(PEMElectrolyzerData, self).build()

        self.electricity_to_mol = Var(self.flowsheet().config.time,
                                      domain=Reals,
                                      initialize=0.0,
                                      doc="Efficiency",
                                      units=pyunits.mol/pyunits.kW/pyunits.second)

        self.electricity = Var(self.flowsheet().config.time,
                               domain=Reals,
                               initialize=0.0,
                               doc="Electricity into control volume",
                               units=pyunits.kW)

        self.electricity_in = Port(noruleinit=True, doc="A port for electricity flow")
        self.electricity_in.add(self.electricity, "electricity")

        self.outlet_state = self.config.property_package.build_state_block(self.flowsheet().config.time,
                                                                           **self.config.property_package_args)
        self.add_outlet_port(name="outlet",
                             block=self.outlet_state,
                             doc="H2 out of electrolyzer")

        @self.Constraint(self.flowsheet().config.time)
        def efficiency_curve(b, t):
            return pyunits.convert(b.outlet.flow_mol[t], to_units=pyunits.mol / pyunits.s) == b.electricity[t] * \
                   b.electricity_to_mol[t]

    def _get_performance_contents(self, time_point=0):
        return {"vars": {"Efficiency": self.electricity_to_mol[time_point]}}

    def initialize_build(self, solver=None, optarg=None, outlvl=idaeslog.NOTSET, **kwargs):
        for t in self.flowsheet().config.time:
            calculate_variable_from_constraint(self.outlet.flow_mol[t],
                                               self.efficiency_curve[t])

        self.outlet_state.initialize(hold_state=False,
                                     solver=solver,
                                     optarg=optarg,
                                     outlvl=outlvl)

    def report(self, time_point=0, dof=False, ostream=None, prefix=""):
        time_point = float(time_point)

        if ostream is None:
            ostream = sys.stdout

        # Get DoF and model stats
        if dof:
            dof_stat = degrees_of_freedom(self)
            nv = number_variables(self)
            nc = number_activated_constraints(self)
            nb = number_activated_blocks(self)

        # Get stream table
        stream_table = create_stream_table_dataframe({"Outlet": self.outlet}, time_point=time_point)
        stream_table.loc['Electricity'] = pd.Series({'Outlet': '-'})
        stream_table.insert(0, "Inlet", ["-", "-", "-", "-", value(self.electricity[time_point])])

        # Set model type output
        if hasattr(self, "is_flowsheet") and self.is_flowsheet:
            model_type = "Flowsheet"
        else:
            model_type = "Unit"

        # Write output
        max_str_length = 84
        tab = " " * 4
        ostream.write("\n" + "=" * max_str_length + "\n")

        lead_str = f"{prefix}{model_type} : {self.name}"
        trail_str = f"Time: {time_point}"
        mid_str = " " * (max_str_length - len(lead_str) - len(trail_str))
        ostream.write(lead_str + mid_str + trail_str)

        if dof:
            ostream.write("\n" + "=" * max_str_length + "\n")
            ostream.write(f"{prefix}{tab}Local Degrees of Freedom: {dof_stat}")
            ostream.write('\n')
            ostream.write(f"{prefix}{tab}Total Variables: {nv}{tab}"
                          f"Activated Constraints: {nc}{tab}"
                          f"Activated Blocks: {nb}")

        if stream_table is not None:
            ostream.write("\n" + "-" * max_str_length + "\n")
            ostream.write(f"{prefix}{tab}Stream Table")
            ostream.write('\n')
            ostream.write(
                textwrap.indent(
                    stream_table_dataframe_to_string(stream_table),
                    prefix + tab))
        ostream.write("\n" + "=" * max_str_length + "\n")



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
# Import objects from pyomo package
from pyomo.environ import ConcreteModel, SolverFactory, Var, TerminationCondition, SolverStatus

# Import the main FlowsheetBlock from IDAES. The flowsheet block will contain the unit model
from idaes.core import FlowsheetBlock
from idaes.core.util.testing import initialization_tester

# Import the H2 property package to create a properties block for the flowsheet
from idaes.models.properties.modular_properties.base.generic_property \
    import GenericParameterBlock

from dispatches.properties.h2_ideal_vap import configuration
from dispatches.unit_models import PEM_Electrolyzer


def test_pem():
    # Create the ConcreteModel and the FlowsheetBlock, and attach the flowsheet block to it.
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)   # dynamic or ss flowsheet needs to be specified here

    # Add properties parameter block to the flowsheet with specifications
    m.fs.properties = GenericParameterBlock(**configuration)

    m.fs.unit = PEM_Electrolyzer(property_package=m.fs.properties)

    assert hasattr(m.fs.unit, "efficiency_curve")
    assert hasattr(m.fs.unit, "electricity_in")
    assert hasattr(m.fs.unit, "outlet")
    assert hasattr(m.fs.unit, "outlet_state")
    assert isinstance(m.fs.unit.electricity, Var)
    assert isinstance(m.fs.unit.electricity_to_mol, Var)

    m.fs.unit.electricity_in.electricity.fix(1)
    m.fs.unit.electricity_to_mol.fix(5)
    initialization_tester(m)

    solver = SolverFactory('ipopt')
    results = solver.solve(m.fs)

    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    assert m.fs.unit.electricity_in.electricity[0].value == 1
    assert m.fs.unit.outlet.flow_mol[0].value == 5.0
    assert m.fs.unit.outlet.temperature[0].value == 300
    assert m.fs.unit.outlet.pressure[0].value == 100000

    m.fs.unit.report(dof=True)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
"""
Tests for ConcreteTubeSide model.

Author: Konica Mulani, Jaffer Ghouse
"""

import pytest
from pyomo.environ import (ConcreteModel, TerminationCondition,
                           SolverStatus, value, units as pyunits)

from idaes.core import FlowsheetBlock
from idaes.models.unit_models.heat_exchanger \
    import HeatExchangerFlowPattern

from idaes.models.properties import iapws95

from idaes.core.util.model_statistics import degrees_of_freedom

from idaes.core.solvers import get_solver
from idaes.core.util.testing import initialization_tester
from dispatches.unit_models import ConcreteTubeSide

# Get default solver for testing
solver = get_solver()


class TestConcreteTube(object):

    "Test for concrete tube side"

    @pytest.fixture(scope="class")
    def concrete_tube(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = iapws95.Iapws95ParameterBlock(
            phase_presentation=iapws95.PhaseType.LG,
        )

        m.fs.unit = ConcreteTubeSide(
            property_package=m.fs.properties,
            flow_type=HeatExchangerFlowPattern.cocurrent,
        )

        m.fs.unit.d_tube_outer.fix(0.01167)
        m.fs.unit.d_tube_inner.fix(0.01167)
        m.fs.unit.tube_length.fix(4.85)
        m.fs.unit.tube_heat_transfer_coefficient.fix(500)

        m.fs.unit.tube_inlet.flow_mol[0].fix(1)  # mol/s
        m.fs.unit.tube_inlet.pressure[0].fix(101325)  # Pa
        m.fs.unit.tube_inlet.enth_mol[0].\
            fix(iapws95.htpx(300*pyunits.K, 101325*pyunits.Pa))  # K

        m.fs.unit.temperature_wall[0, :].fix(1000)

        return m

    @pytest.mark.unit
    def test_build(self, concrete_tube):

        assert len(concrete_tube.fs.unit.config) == 14
        assert concrete_tube.fs.unit.config.flow_type == \
            HeatExchangerFlowPattern.cocurrent
        assert concrete_tube.fs.unit.config.finite_elements == 20
        assert concrete_tube.fs.unit.config.collocation_points == 5

        assert hasattr(concrete_tube.fs.unit, "tube_area")
        assert hasattr(concrete_tube.fs.unit, "tube_length")
        assert hasattr(concrete_tube.fs.unit, "d_tube_outer")
        assert hasattr(concrete_tube.fs.unit, "d_tube_inner")
        assert hasattr(concrete_tube.fs.unit,
                       "tube_heat_transfer_coefficient")
        assert hasattr(concrete_tube.fs.unit, "temperature_wall")
        assert hasattr(concrete_tube.fs.unit, "tube_heat_transfer_eq")
        assert hasattr(concrete_tube.fs.unit, "area_calc_tube")

        assert degrees_of_freedom(concrete_tube) == 0

    @pytest.mark.component
    def test_initialize(self, concrete_tube):
        initialization_tester(concrete_tube)

    @pytest.mark.component
    def test_solve(self, concrete_tube):
        results = solver.solve(concrete_tube)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

        assert pytest.approx(1, abs=1e-5) == \
            value(concrete_tube.fs.unit.tube_outlet.flow_mol[0])

        assert pytest.approx(55702.16, abs=1e0) == \
            value(concrete_tube.fs.unit.tube_outlet.enth_mol[0])

        assert pytest.approx(101325, abs=1e2) == \
            value(concrete_tube.fs.unit.tube_outlet.pressure[0])

        assert pytest.approx(581.80, abs=1e2) == \
            value(concrete_tube.fs.unit.tube.properties[0, 1].temperature)

    @pytest.mark.component
    def test_conservation(self, concrete_tube):
        assert abs(value(concrete_tube.fs.unit.tube_inlet.flow_mol[0] -
                         concrete_tube.fs.unit.tube_outlet.flow_mol[0])) \
            <= 1e-6

        tube_side = value(pyunits.convert(
                concrete_tube.fs.unit.tube_outlet.flow_mol[0]
                * (concrete_tube.fs.unit.tube.
                   properties[0, 1].enth_mol_phase['Liq']
                   - concrete_tube.fs.unit.tube.properties[0, 0].
                   enth_mol_phase['Liq']),
                to_units=pyunits.W))
        # NS: updated the assert statment with a new computed value due changes in property package
        # assert abs(tube_side) == pytest.approx(23497.05, rel=1e-3)
        assert abs(tube_side) == pytest.approx(23091.038, rel=1e-1)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
"""
Renewable Energy Flowsheet
Author: Darice Guittet
Date: June 7, 2021
"""

import matplotlib.pyplot as plt
from functools import partial
from pyomo.environ import (Constraint,
                           Var,
                           ConcreteModel,
                           Expression,
                           Param,
                           units as pyunits,
                           SolverFactory,
                           TransformationFactory,
                           NonNegativeReals,
                           Reference,
                           value)
from pyomo.network import Arc, Port
import idaes.core.util.scaling as iscale

import idaes.logger as idaeslog
from idaes.core import FlowsheetBlock
from idaes.core.util.initialization import propagate_state
from idaes.models.properties.modular_properties.base.generic_property \
    import GenericParameterBlock
from idaes.models.unit_models import (Translator,
                                              Mixer,
                                              MomentumMixingType,
                                              Valve,
                                              ValveFunctionType)

from dispatches.properties.h2_ideal_vap import configuration as h2_ideal_config
from dispatches.properties.hturbine_ideal_vap import configuration as hturbine_config
import dispatches.properties.h2_reaction as h2_reaction_props
from dispatches.unit_models import (
    HydrogenTurbine,
    HydrogenTank,
    SimpleHydrogenTank,
    PEM_Electrolyzer,
    ElectricalSplitter,
    BatteryStorage,
    Wind_Power
)

from idaes.models.unit_models.separator import Separator
from dispatches.case_studies.renewables_case.load_parameters import *
import matplotlib as mpl
mpl.use(mpl.rcParams['backend'])


def add_wind(m, wind_mw, wind_resource_config=None):
    """
    Adds a wind unit to the flowsheet with a fixed system capacity and that uses resource data to determine the capacity factors

    The resource data can be provided by configuring either `resource_speed` or `resource_probability_density`.
    The `resource_speed` is a list of wind speeds (m/s) per timestep of the flowsheet.
    The `resource_probability_density` is formatted as a 2d list where each row contains: (wind m/s, wind degrees from north clockwise, probability) 
    and the probabilities in all rows add up to 1.

    Args:
        m: existing ConcreteModel with a flowsheet `fs`
        wind_mw: capacity of wind model to be fixed
        wind_resource_config: dictionary of Windpower Config keys (`resource_speed`, `resource_probability_density`) and ConfigValues
    Returns:
        wind unit model in the flowsheet
    """
    m.fs.windpower = Wind_Power(**wind_resource_config)
    m.fs.windpower.system_capacity.fix(wind_mw * 1e3)   # kW
    return m.fs.windpower


def add_pem(m, outlet_pressure_bar):
    """
    Adds a PEM electrolyzer unit to the flowsheet with a fixed outlet pressure and conversion rate. The properties package is h2_ideal_vap.

    The operating temperature and maximum pressure of the PEM are pulled from load_parameters.py

    Args:
        m: existing ConcreteModel with a flowsheet `fs`
        outlet_pressure_bar: operating pressure of the PEM
    Returns:
        PEM electrolyzer unit model in the flowsheet
        h2_ideal_vap property package in the flowsheet

    """
    m.fs.h2ideal_props = GenericParameterBlock(**h2_ideal_config)
    m.fs.h2ideal_props.set_default_scaling('flow_mol_phase', 1)
    m.fs.h2ideal_props.set_default_scaling('mole_frac_comp', 1)
    m.fs.h2ideal_props.set_default_scaling('mole_frac_phase_comp', 1)
    m.fs.h2ideal_props.set_default_scaling('flow_mol', 1)
    m.fs.h2ideal_props.set_default_scaling('enth_mol_phase', 0.1)

    m.fs.pem = PEM_Electrolyzer(property_package=m.fs.h2ideal_props)

    # Conversion of kW to mol/sec of H2. (elec*elec_to_mol) based on H-tec design of 54.517kW-hr/kg
    m.fs.pem.electricity_to_mol.fix(0.002527406)
    m.fs.pem.outlet.pressure.setub(max_pressure_bar * 1e5)
    m.fs.pem.outlet.pressure.fix(outlet_pressure_bar * 1e5)
    m.fs.pem.outlet.temperature.fix(pem_temp)
    return m.fs.pem, m.fs.h2ideal_props


def add_battery(m, batt_mw):
    """
    Adds a Battery unit to the flowsheet with a fixed power and duration of 4 hours.
    
    The charging and discharging efficiencies are set at 98%.

    Args:
        m: existing ConcreteModel with a flowsheet `fs`
        batt_mw: nameplate power
    Returns:
        battery unit model in flowsheet
    """
    m.fs.battery = BatteryStorage()
    m.fs.battery.charging_eta.set_value(0.95)
    m.fs.battery.discharging_eta.set_value(0.95)
    m.fs.battery.dt.set_value(timestep_hrs)
    m.fs.battery.nameplate_power.fix(batt_mw * 1e3)
    m.fs.battery.duration = Param(default=4, mutable=True, units=pyunits.kWh/pyunits.kW)
    m.fs.battery.four_hr_battery = Constraint(expr=m.fs.battery.nameplate_power * m.fs.battery.duration == m.fs.battery.nameplate_energy)
    return m.fs.battery


def add_h2_tank(m, tank_type="simple", valve_outlet_bar=None, length_m=None):
    """
    Adds a Hydrogen Tank unit to the flowsheet, with 3 options for the model type: `simple`, `detailed` and `detailed-valve`.

    For the `simple` type, the model is linear with no energy balances.
    For the `detailed` type, the model is nonlinear with energy balances deactivated. The capacity of the tank is determined by its 
    geometry.
    For the `detailed-valve` type, the `detailed` model is used, but with energy balances active and with a Valve unit. 
    The valve coefficient is pulled from load_parameters but needs to be be tuned per application.

    Args:
        m: existing ConcreteModel with a flowsheet `fs`
        tank_type: `simple`, `detailed`, or `detailed-valve`
        valve_outlet_bar: required if not using `simple` type, outlet pressure of the valve
        length_m: required if using `detailed` type, length of tank
    Returns:
        tank (and valve) unit model(s) in flowsheet
    """
    if "detailed" in tank_type:
        m.fs.h2_tank = HydrogenTank(property_package=m.fs.h2ideal_props, dynamic=False)
        m.fs.h2_tank.tank_diameter.fix(0.1)
        m.fs.h2_tank.tank_length.fix(length_m)
        m.fs.h2_tank.control_volume.properties_in[0].pressure.setub(max_pressure_bar * 1e5)
        m.fs.h2_tank.control_volume.properties_out[0].pressure.setub(max_pressure_bar * 1e5)
        m.fs.h2_tank.previous_state[0].pressure.setub(max_pressure_bar * 1e5)
        if tank_type == "detailed-valve":
            m.fs.tank_valve = Valve(
                valve_function_callback=ValveFunctionType.linear,
                property_package=m.fs.h2ideal_props,
            )
            m.fs.tank_to_valve = Arc(
                source=m.fs.h2_tank.outlet,
                destination=m.fs.tank_valve.inlet
            )
            m.fs.tank_valve.outlet.pressure[0].fix(valve_outlet_bar * 1e5)
            # tune valve's coefficient of flow to match the condition
            m.fs.tank_valve.Cv.fix(valve_cv)
            # unfixing valve opening. This allows for controlling both pressure and flow at the outlet of the valve
            m.fs.tank_valve.valve_opening[0].unfix()
            m.fs.tank_valve.valve_opening[0].setlb(0)
        else:
            m.fs.h2_tank.energy_balances.deactivate()
    elif tank_type == "simple":
        m.fs.h2_tank = SimpleHydrogenTank(property_package=m.fs.h2ideal_props, dynamic=False)
        m.fs.h2_tank.outlet_to_turbine.mole_frac_comp[0, "hydrogen"].fix(1)
        m.fs.h2_tank.outlet_to_pipeline.mole_frac_comp[0, "hydrogen"].fix(1)
    else:
        raise ValueError(f"Unrecognized tank_type {tank_type}")

    m.fs.h2_tank.dt[0].fix(timestep_hrs * 3600)

    return m.fs.h2_tank


def add_h2_turbine(m, inlet_pres_bar):
    """
    Adds a hydrogen turbine unit to the flowsheet, using the h2_reaction_props property package.

    A translator and mixer unit is added also. The mixer adds an `air_feed` and an `purchased_hydrogen_feed` to the `hydrogen_feed`.
    The air_feed contains oxygen at a fixed ratio relative to the inlet hydrogen flow rate, `air_h2_ratio`, from load_parameters.
    .The purchased_hydrogen_feed allows the turbine hydrogen inlet flow rate to be `h2_turb_min_flow`, which helps
    with the model solve.

    The `compressor_dp` is fixed from load_parameters, the compressor's isentropic efficiency is 0.86, 
    the stoichiometric conversion rate of hydrogen is 0.99, and the turbine's isentropic efficiency is 0.89.

    The turbine's `electricity` is an Expression taking the difference between the turbine's and compressor's work.

    Args:
        m: existing ConcreteModel with a flowsheet `fs`
        inlet_pres_bar: operating pressure of inlet air and hydrogen to the mixer
    Returns:
        tank (and valve, if applicable) unit model(s) in flowsheet
    """
    m.fs.h2turbine_props = GenericParameterBlock(**hturbine_config)

    m.fs.reaction_params = h2_reaction_props.H2ReactionParameterBlock(
        property_package=m.fs.h2turbine_props,
    )

    # Add translator block
    m.fs.translator = Translator(
        inlet_property_package=m.fs.h2ideal_props,
        outlet_property_package=m.fs.h2turbine_props,
    )

    m.fs.translator.eq_flow_hydrogen = Constraint(
        expr=m.fs.translator.inlet.flow_mol[0] ==
        m.fs.translator.outlet.flow_mol[0]
    )

    m.fs.translator.eq_temperature = Constraint(
        expr=m.fs.translator.inlet.temperature[0] ==
        m.fs.translator.outlet.temperature[0]
    )

    m.fs.translator.eq_pressure = Constraint(
        expr=m.fs.translator.inlet.pressure[0] ==
        m.fs.translator.outlet.pressure[0]
    )

    m.fs.translator.outlet.mole_frac_comp[0, "hydrogen"].fix(0.99)
    m.fs.translator.outlet.mole_frac_comp[0, "oxygen"].fix(0.01/4)
    m.fs.translator.outlet.mole_frac_comp[0, "argon"].fix(0.01/4)
    m.fs.translator.outlet.mole_frac_comp[0, "nitrogen"].fix(0.01/4)
    m.fs.translator.outlet.mole_frac_comp[0, "water"].fix(0.01/4)

    m.fs.translator.inlet.pressure[0].setub(max_pressure_bar * 1e5)
    m.fs.translator.outlet.pressure[0].setub(max_pressure_bar * 1e5)

    # Add mixer block
    # purchased_hydrogen_feed as slack for turbine inlet flow mol to be nonzero
    m.fs.mixer = Mixer(
        # using minimize pressure for all inlets and outlet of the mixer
        # because pressure of inlets is already fixed in flowsheet, using equality will over-constrain
        momentum_mixing_type=MomentumMixingType.minimize,
        property_package=m.fs.h2turbine_props,
        inlet_list=["air_feed", "hydrogen_feed", "purchased_hydrogen_feed"],
    )

    m.fs.mixer.air_feed.temperature[0].fix(pem_temp)
    m.fs.mixer.air_feed.pressure[0].fix(inlet_pres_bar * 1e5)
    m.fs.mixer.air_feed.mole_frac_comp[0, "oxygen"].fix(0.2054)
    m.fs.mixer.air_feed.mole_frac_comp[0, "argon"].fix(0.0032)
    m.fs.mixer.air_feed.mole_frac_comp[0, "nitrogen"].fix(0.7672)
    m.fs.mixer.air_feed.mole_frac_comp[0, "water"].fix(0.0240)
    m.fs.mixer.air_feed.mole_frac_comp[0, "hydrogen"].fix(2e-4)
    m.fs.mixer.purchased_hydrogen_feed.pressure[0].fix(inlet_pres_bar * 1e5)
    m.fs.mixer.purchased_hydrogen_feed.temperature[0].fix(pem_temp)
    m.fs.mixer.purchased_hydrogen_feed.mole_frac_comp[0, "hydrogen"].fix(0.99)
    m.fs.mixer.purchased_hydrogen_feed.mole_frac_comp[0, "oxygen"].fix(0.01/4)
    m.fs.mixer.purchased_hydrogen_feed.mole_frac_comp[0, "argon"].fix(0.01/4)
    m.fs.mixer.purchased_hydrogen_feed.mole_frac_comp[0, "nitrogen"].fix(0.01/4)
    m.fs.mixer.purchased_hydrogen_feed.mole_frac_comp[0, "water"].fix(0.01/4)

    m.fs.mixer.mixed_state[0].pressure.setub(max_pressure_bar * 1e5)
    m.fs.mixer.air_feed_state[0].pressure.setub(max_pressure_bar * 1e5)
    m.fs.mixer.hydrogen_feed_state[0].pressure.setub(max_pressure_bar * 1e5)
    m.fs.mixer.purchased_hydrogen_feed_state[0].pressure.setub(max_pressure_bar * 1e5)

    m.fs.mixer.air_h2_ratio = Constraint(
        expr=m.fs.mixer.air_feed.flow_mol[0] == air_h2_ratio * (
                m.fs.mixer.purchased_hydrogen_feed.flow_mol[0] + m.fs.mixer.hydrogen_feed.flow_mol[0]))

    m.fs.mixer.purchased_hydrogen_feed.flow_mol[0].setlb(h2_turb_min_flow / 2)

    m.fs.translator_to_mixer = Arc(
        source=m.fs.translator.outlet,
        destination=m.fs.mixer.hydrogen_feed
    )

    m.fs.h2_turbine = HydrogenTurbine(
        property_package=m.fs.h2turbine_props,
        reaction_package=m.fs.reaction_params,
    )
    m.fs.h2_turbine.compressor.deltaP.fix(compressor_dp * 1e5)
    m.fs.h2_turbine.compressor.efficiency_isentropic.fix(0.86)

    # Specify the Stoichiometric Conversion Rate of hydrogen
    # in the equation shown below
    # H2(g) + O2(g) --> H2O(g) + energy
    m.fs.h2_turbine.stoic_reactor.conversion.fix(0.99)

    m.fs.h2_turbine.turbine.deltaP.fix(-compressor_dp * 1e5)
    m.fs.h2_turbine.turbine.efficiency_isentropic.fix(0.89)

    # add operating constraints
    m.fs.h2_turbine.electricity = Expression(m.fs.config.time,
                                             rule=lambda b, t: (-b.turbine.work_mechanical[t] - b.compressor.work_mechanical[t]) * 1e-3)

    m.fs.mixer_to_turbine = Arc(
        source=m.fs.mixer.outlet,
        destination=m.fs.h2_turbine.compressor.inlet
    )

    return m.fs.h2_turbine, m.fs.mixer, m.fs.translator


def create_model(wind_mw, pem_bar, batt_mw, tank_type, tank_length_m, turb_inlet_bar, wind_resource_config=None):
    """
    Creates a Flowsheet Pyomo model that puts together the Wind unit model with optional PEM, Hydrogen Tank, and Hydrogen Turbine unit models.

    The input parameters determine the size of the technologies by fixing the appropriate variable. If the size parameter is None, the technology
    will not be added.

    The wind is first split among its output destinations: grid, battery and PEM with an ElectricalSplitter unit model. 
    After the PEM, a tank and turbine may be added. The `simple` tank model includes outlet ports for hydrogen flows to the turbine and the pipeline.
    The `detailed` tank model uses a Splitter unit model to split the material and energy flows to the turbine and to hydrogen sales.

    Args:
        wind_mw: wind farm capacity
        pem_bar: operating pressure of the PEM  
        batt_mw: nameplate power
        tank_type: `simple`, `detailed`, or `detailed-valve`. See `add_h2_tank` for descriptions
        tank_length_m: required if using `detailed` tank type, length of tank
        turb_inlet_bar: operating inlet pressure of hydrogen to turbine
        wind_resource_config: dictionary of Windpower Config keys (`resource_speed`, `resource_probability_density`) and ConfigValues. See `add_wind` for description
    
    """
    m = ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    wind = add_wind(m, wind_mw, wind_resource_config)
    wind_output_dests = ["grid"]

    if pem_bar is not None:
        pem, _ = add_pem(m, pem_bar)
        wind_output_dests.append("pem")

    if batt_mw is not None:
        battery = add_battery(m, batt_mw)
        wind_output_dests.append("battery")

    if tank_length_m is not None:
        h2_tank = add_h2_tank(m, tank_type, pem_bar, tank_length_m)

    if turb_inlet_bar is not None and tank_length_m is not None:
        add_h2_turbine(m, turb_inlet_bar)

    # Set up where wind output flows to
    if len(wind_output_dests) > 1:
        m.fs.splitter = ElectricalSplitter(outlet_list=wind_output_dests)
        m.fs.wind_to_splitter = Arc(source=wind.electricity_out, dest=m.fs.splitter.electricity_in)

    if "pem" in wind_output_dests:
        m.fs.splitter_to_pem = Arc(source=m.fs.splitter.pem_port, dest=pem.electricity_in)
    if "battery" in wind_output_dests:
        m.fs.splitter_to_battery = Arc(source=m.fs.splitter.battery_port, dest=battery.power_in)

    if hasattr(m.fs, "h2_tank"):
        m.fs.pem_to_tank = Arc(source=pem.outlet, dest=h2_tank.inlet)

    if hasattr(m.fs, "h2_turbine"):
        if tank_type == 'simple':
            m.fs.h2_tank_to_turb = Arc(source=m.fs.h2_tank.outlet_to_turbine,
                                       destination=m.fs.translator.inlet)
        else:
            m.fs.h2_splitter = Separator(
                property_package=m.fs.h2ideal_props,
                outlet_list=["sold", "turbine"],
            )
            if tank_type == 'detailed-valve':
                m.fs.valve_to_h2_splitter = Arc(source=m.fs.tank_valve.outlet,
                                                destination=m.fs.h2_splitter.inlet)
            else:
                m.fs.valve_to_h2_splitter = Arc(source=m.fs.h2_tank.outlet,
                                                destination=m.fs.h2_splitter.inlet)
            # Set up where hydrogen from tank flows to
            m.fs.h2_splitter_to_turb = Arc(source=m.fs.h2_splitter.turbine,
                                        destination=m.fs.translator.inlet)

    TransformationFactory("network.expand_arcs").apply_to(m)

    # Scaling factors, set mostly to 1 for now
    elec_sf = 1
    iscale.set_scaling_factor(m.fs.windpower.electricity, elec_sf)
    if hasattr(m.fs, "splitter"):
        iscale.set_scaling_factor(m.fs.splitter.electricity, elec_sf)
        iscale.set_scaling_factor(m.fs.splitter.grid_elec, elec_sf)

    if hasattr(m.fs, "battery"):
        iscale.set_scaling_factor(m.fs.splitter.battery_elec, elec_sf)
        iscale.set_scaling_factor(m.fs.battery.elec_in, elec_sf)
        iscale.set_scaling_factor(m.fs.battery.elec_out, elec_sf)
        iscale.set_scaling_factor(m.fs.battery.nameplate_power, elec_sf)
        iscale.set_scaling_factor(m.fs.battery.nameplate_energy, elec_sf)
        iscale.set_scaling_factor(m.fs.battery.initial_state_of_charge, elec_sf)
        iscale.set_scaling_factor(m.fs.battery.initial_energy_throughput, elec_sf)
        iscale.set_scaling_factor(m.fs.battery.state_of_charge, elec_sf)

    if hasattr(m.fs, "pem"):
        iscale.set_scaling_factor(m.fs.splitter.pem_elec, elec_sf)
        iscale.set_scaling_factor(m.fs.pem.electricity, elec_sf)

    if hasattr(m.fs, "h2_turbine"):
        iscale.set_scaling_factor(m.fs.mixer.minimum_pressure, 1)
        iscale.set_scaling_factor(m.fs.mixer.air_feed_state[0.0].enth_mol_phase['Vap'], 1)
        iscale.set_scaling_factor(m.fs.mixer.hydrogen_feed_state[0.0].enth_mol_phase['Vap'], 1)
        iscale.set_scaling_factor(m.fs.mixer.purchased_hydrogen_feed_state[0.0].enth_mol_phase['Vap'], 1)
        iscale.set_scaling_factor(m.fs.mixer.mixed_state[0.0].enth_mol_phase['Vap'], 1)
        iscale.set_scaling_factor(m.fs.mixer.hydrogen_feed_state[0.0].flow_mol_phase['Vap'], 1)

        iscale.set_scaling_factor(m.fs.h2_turbine.compressor.control_volume.properties_in[0.0].enth_mol_phase['Vap'], 1)
        iscale.set_scaling_factor(m.fs.h2_turbine.compressor.control_volume.properties_out[0.0].enth_mol_phase['Vap'], 1)
        iscale.set_scaling_factor(m.fs.h2_turbine.compressor.control_volume.work, 1)
        iscale.set_scaling_factor(m.fs.h2_turbine.compressor.properties_isentropic[0.0].enth_mol_phase['Vap'], 1)
        iscale.set_scaling_factor(m.fs.h2_turbine.stoic_reactor.control_volume.properties_in[0.0].enth_mol_phase['Vap'], 1)
        iscale.set_scaling_factor(m.fs.h2_turbine.stoic_reactor.control_volume.properties_out[0.0].enth_mol_phase['Vap'], 1)
        iscale.set_scaling_factor(m.fs.h2_turbine.stoic_reactor.control_volume.rate_reaction_extent[0, 'R1'], 1)
        iscale.set_scaling_factor(m.fs.h2_turbine.turbine.control_volume.properties_in[0.0].enth_mol_phase['Vap'], 1)
        iscale.set_scaling_factor(m.fs.h2_turbine.turbine.control_volume.properties_out[0.0].enth_mol_phase['Vap'], 1)
        iscale.set_scaling_factor(m.fs.h2_turbine.turbine.control_volume.work, 1)
        iscale.set_scaling_factor(m.fs.h2_turbine.turbine.properties_isentropic[0.0].enth_mol_phase['Vap'], 1)

    iscale.calculate_scaling_factors(m)
    return m



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
import pytest
import platform
from idaes.core.util.model_statistics import degrees_of_freedom

from dispatches.unit_models import HydrogenTank as DetailedHydrogenTank
from dispatches.case_studies.renewables_case.RE_flowsheet import *
from dispatches.case_studies.renewables_case.wind_battery_LMP import wind_battery_optimize, record_results, plot_results
from dispatches.case_studies.renewables_case.wind_battery_PEM_LMP import wind_battery_pem_optimize
from dispatches.case_studies.renewables_case.wind_battery_PEM_tank_turbine_LMP import wind_battery_pem_tank_turb_optimize

@pytest.fixture
def input_params():
    params = copy.copy(default_input_params)
    with open(re_case_dir / 'tests' / 'rts_results_all_prices.npy', 'rb') as f:
        _ = np.load(f)
        price = np.load(f)

        prices_used = copy.copy(price)
        prices_used[prices_used > 200] = 200
    params['DA_LMPs'] = prices_used

    # wind resource data from example Wind Toolkit file
    wind_data = SRW_to_wind_data(re_case_dir / 'data' / '44.21_-101.94_windtoolkit_2012_60min_80m.srw')
    wind_speeds = [wind_data['data'][i][2] for i in range(8760)]
    wind_resource = {t:
                        {'wind_resource_config': {
                            'resource_speed': [wind_speeds[t]]
                        }
                    } for t in range(8760)}
    params["wind_resource"] = wind_resource
    return params

def test_h2_valve_opening():
    valve_coef = 0.03380
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.h2ideal_props = GenericParameterBlock(**h2_ideal_config)

    m.fs.h2_tank = DetailedHydrogenTank(property_package=m.fs.h2ideal_props, dynamic=False)
    m.fs.h2_tank.tank_diameter.fix(0.1)
    m.fs.h2_tank.tank_length.fix(0.3)
    m.fs.h2_tank.control_volume.properties_in[0].pressure.setub(max_pressure_bar * 1e5)
    m.fs.h2_tank.control_volume.properties_out[0].pressure.setub(max_pressure_bar * 1e5)
    m.fs.h2_tank.previous_state[0].pressure.setub(max_pressure_bar * 1e5)
    # hydrogen tank valve
    m.fs.tank_valve = Valve(
        valve_function_callback=ValveFunctionType.linear,
        property_package=m.fs.h2ideal_props,
    )
    # connect tank to the valve
    m.fs.tank_to_valve = Arc(
        source=m.fs.h2_tank.outlet,
        destination=m.fs.tank_valve.inlet
    )
    m.fs.tank_valve.outlet.pressure[0].fix(3 * 1e5)
    # NS: tuning valve's coefficient of flow to match the condition
    m.fs.tank_valve.Cv.fix(valve_coef)
    # NS: unfixing valve opening. This allows for controlling both pressure
    # and flow at the outlet of the valve
    m.fs.tank_valve.valve_opening[0].unfix()
    m.fs.tank_valve.valve_opening[0].setlb(0)
    m.fs.h2_tank.dt[0].fix(timestep_hrs * 3600)

    assert m.fs.h2_tank.tank_diameter[0].fixed
    assert m.fs.h2_tank.tank_length[0].fixed
    assert m.fs.tank_valve.outlet.pressure[0].fixed

    m.fs.h2_tank.inlet.flow_mol[0].fix(9992.784)
    m.fs.h2_tank.outlet.flow_mol[0].fix(4251.928)

    m.fs.tank_valve.valve_opening[0].setlb(0)
    m.fs.tank_valve.valve_opening[0].setub(1)

    m.fs.h2_tank.previous_material_holdup[(0.0, 'Vap', 'hydrogen')].fix(17.832)
    dof = degrees_of_freedom(m)
    assert dof == 8

    opt = SolverFactory("ipopt")
    res = opt.solve(m, tee=True)

    assert res.Solver.status == 'ok'
    assert value(m.fs.h2_tank.material_holdup[(0.0, 'Vap', 'hydrogen')]) == pytest.approx(22.3761, rel=1e-3)
    assert value(m.fs.tank_valve.valve_opening[0]) == pytest.approx(0.83331, rel=1e-3)


def test_create_model(input_params):
    tank_type = "simple"
    m = create_model(
        wind_mw=fixed_wind_mw,
        pem_bar=pem_bar,
        batt_mw=fixed_batt_mw,
        tank_type=tank_type,
        tank_length_m=fixed_tank_size,
        turb_inlet_bar=pem_bar,
        wind_resource_config=input_params['wind_resource'][0]['wind_resource_config']
    )

    assert hasattr(m.fs, "windpower")
    assert hasattr(m.fs, "splitter")
    assert hasattr(m.fs, "battery")
    assert hasattr(m.fs, "pem")
    assert hasattr(m.fs, "h2_tank")
    assert hasattr(m.fs, "translator")
    if tank_type != "simple":
        assert hasattr(m.fs, "h2_splitter")
        assert hasattr(m.fs, "tank_sold")
    assert hasattr(m.fs, "mixer")
    assert hasattr(m.fs, "h2_turbine")
    assert hasattr(m.fs.mixer, "purchased_hydrogen_feed")

    assert m.fs.windpower.system_capacity.fixed
    assert m.fs.battery.nameplate_power.fixed
    if tank_type != "simple":
        assert m.fs.h2_tank.tank_length[0].fixed
        assert m.fs.h2_tank.energy_balances.active
    assert m.fs.mixer.air_h2_ratio.active
    assert m.fs.mixer.purchased_hydrogen_feed.flow_mol[0].lb
    assert value(m.fs.h2_turbine.turbine.deltaP[0]) == -2401000.0
    assert value(m.fs.mixer.air_feed.mole_frac_comp[0, "hydrogen"]) == 2e-4

    dof = degrees_of_freedom(m)
    assert dof == 10


def test_wind_battery_optimize(input_params):
    mp = wind_battery_optimize(n_time_points=7 * 24, input_params=input_params, verbose=True)
    assert value(mp.pyomo_model.NPV) == pytest.approx(1001068228, rel=1e-3)
    assert value(mp.pyomo_model.annual_revenue) == pytest.approx(168691601, rel=1e-3)
    blks = mp.get_active_process_blocks()
    assert value(blks[0].fs.battery.nameplate_power) == pytest.approx(1326779, rel=1e-3)
    plot_results(*record_results(mp))


def test_wind_battery_pem_optimize(input_params):
    input_params['h2_price_per_kg'] = 2.5
    design_res, _ = wind_battery_pem_optimize(time_points=6 * 24, input_params=input_params, verbose=True)
    assert design_res['batt_mw'] == pytest.approx(4874, rel=1e-3)
    assert design_res['pem_mw'] == pytest.approx(0, abs=1e-1)
    assert design_res['annual_rev_h2'] == pytest.approx(0.055, abs=1e-1)
    assert design_res['annual_rev_E'] == pytest.approx(531576401, rel=1e-2)
    assert design_res['NPV'] == pytest.approx(2322131921, rel=1e-2)


def test_wind_battery_pem_tank_turb_optimize_simple(input_params):
    input_params['h2_price_per_kg'] = 2.0
    design_res = wind_battery_pem_tank_turb_optimize(6 * 24, input_params, verbose=False, plot=False)
    assert design_res['batt_mw'] == pytest.approx(4874, rel=1e-2)
    assert design_res['pem_mw'] == pytest.approx(0, abs=3)
    assert design_res['tank_kgH2'] == pytest.approx(0, abs=3)
    assert design_res['turb_mw'] == pytest.approx(0, abs=3)
    assert design_res['avg_turb_eff'] == pytest.approx(1.51, rel=1e-1)
    assert design_res['annual_rev_h2'] == pytest.approx(2634, abs=5e3)
    assert design_res['annual_rev_E'] == pytest.approx(531566543, rel=1e-2)
    assert design_res['NPV'] == pytest.approx(2344545889, rel=1e-2)


@pytest.mark.skipif(platform.system() == "Windows", reason="Platform differences in IPOPT solve")
def test_wind_battery_pem_tank_turb_optimize_detailed(input_params):
    input_params['h2_price_per_kg'] = 2.0
    input_params['tank_type'] = 'detailed'
    design_res = wind_battery_pem_tank_turb_optimize(6 * 24, input_params=input_params, verbose=True, plot=False)
    assert design_res['batt_mw'] == pytest.approx(4874, rel=1e-2)
    assert design_res['pem_mw'] == pytest.approx(0, abs=3)
    assert design_res['tank_kgH2'] == pytest.approx(0, abs=3)
    assert design_res['turb_mw'] == pytest.approx(0, abs=3)
    assert design_res['avg_turb_eff'] == pytest.approx(1.51, rel=1e-1)
    assert design_res['annual_rev_h2'] == pytest.approx(2634, abs=5e3)
    assert design_res['annual_rev_E'] == pytest.approx(531566543, rel=1e-2)
    assert design_res['NPV'] == pytest.approx(2344545889, rel=1e-2)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
from dispatches.case_studies.renewables_case.wind_battery_LMP import (
    wind_battery_mp_block,
    wind_battery_variable_pairs,
    wind_battery_periodic_variable_pairs,
)
from idaes.apps.grid_integration.multiperiod.multiperiod import MultiPeriodModel
import pyomo.environ as pyo
import pandas as pd
from collections import deque
from functools import partial
from dispatches.case_studies.renewables_case.load_parameters import wind_speeds


def create_multiperiod_wind_battery_model(n_time_points, wind_cfs, input_params):
    """This function creates a MultiPeriodModel for the wind battery model.

    Args:
        n_time_points (int): number of time period for the model.

    Returns:
        MultiPeriodModel: a MultiPeriodModel for the wind battery model.
    """

    # create the multiperiod model object
    mp_wind_battery = MultiPeriodModel(
        n_time_points=n_time_points,
        process_model_func=partial(wind_battery_mp_block, input_params=input_params, verbose=False),
        linking_variable_func=wind_battery_variable_pairs,
        periodic_variable_func=wind_battery_periodic_variable_pairs,
    )

    wind_capacity_factors = {t:
                            {'wind_resource_config': {
                                'capacity_factor': 
                                    [wind_cfs[t]]}} for t in range(n_time_points)}

    mp_wind_battery.build_multi_period_model(wind_capacity_factors)

    return mp_wind_battery


def transform_design_model_to_operation_model(
    mp_wind_battery,
    wind_capacity=200e3,
    battery_power_capacity=25e3,
    battery_energy_capacity=100e3,
):
    """Transform the multiperiod wind battery design model to operation model.

    Args:
        mp_wind_battery (MultiPeriodModel): a created multiperiod wind battery object
        wind_capacity (float, optional): wind farm capapcity in KW. Defaults to 200e3.
        battery_power_capacity (float, optional): battery power output capacity in KW. Defaults to 25e3.
        battery_energy_capacity (float, optional): battery energy capacity in KW. Defaults to 100e3.
    """

    blks = mp_wind_battery.get_active_process_blocks()

    for t, b in enumerate(blks):
        b.fs.windpower.system_capacity.fix(wind_capacity)
        b.fs.battery.nameplate_power.fix(battery_power_capacity)
        b.fs.battery.nameplate_energy.fix(battery_energy_capacity)

        if t == 0:
            b.fs.battery.initial_state_of_charge.fix()

        # deactivate periodic boundary condition
        if t == len(blks) - 1:
            b.periodic_constraints[0].deactivate()

    return


def update_wind_capacity_factor(mp_wind_battery, new_capacity_factors):
    """Update the wind capacity factor in the model during rolling horizon.

    Args:
        mp_wind_battery (MultiPeriodModel): a created multiperiod wind battery object
        new_capacity_factors (list): a list of new wind capacity
    """

    blks = mp_wind_battery.get_active_process_blocks()
    for idx, b in enumerate(blks):
        b.fs.windpower.capacity_factor[0] = new_capacity_factors[idx]

    return


class MultiPeriodWindBattery:
    def __init__(
        self,
        model_data,
        wind_capacity_factors=None,
        wind_pmax_mw=200.0,
        battery_pmax_mw=25.0,
        battery_energy_capacity_mwh=100.0,
    ):
        """Initialize a multiperiod wind battery model object for double loop.

        Args:
            model_data (GeneratorModelData): a GeneratorModelData that holds the generators params.
            wind_capacity_factors (list, optional): a list of wind capacity. Defaults to None.
            wind_pmax_mw (float, optional): wind farm capapcity in MW. Defaults to 200.
            battery_pmax_mw (float, optional): battery power output capapcity in MW. Defaults to 25.
            battery_energy_capacity_mwh (float, optional): battery energy capapcity in MW. Defaults to 100.

        Raises:
            ValueError: if wind capacity factor is not provided, ValueError will be raised
        """

        self.model_data = model_data

        if wind_capacity_factors is None:
            raise ValueError("Please provide wind capacity factors.")
        self._wind_capacity_factors = wind_capacity_factors

        self._wind_pmax_mw = wind_pmax_mw
        self._battery_pmax_mw = battery_pmax_mw
        self._battery_energy_capacity_mwh = battery_energy_capacity_mwh

        # a list that holds all the result in pd DataFrame
        self.result_list = []

    def populate_model(self, b, horizon):
        """Create a wind-battery model using the `MultiPeriod` package.

        Args:
            b (block): this is an empty block passed in from eithe a bidder or tracker
            horizon (int): the number of time periods
        """
        blk = b
        if not blk.is_constructed():
            blk.construct()

        input_params = {
            'wind_mw': self._wind_pmax_mw,
            'batt_mw': self._battery_pmax_mw
        }
        blk.windBattery = create_multiperiod_wind_battery_model(horizon, wind_cfs=self._wind_capacity_factors[0:horizon], input_params=input_params)
        transform_design_model_to_operation_model(
            mp_wind_battery=blk.windBattery,
            wind_capacity=self._wind_pmax_mw * 1e3,
            battery_power_capacity=self._battery_pmax_mw * 1e3,
            battery_energy_capacity=self._battery_energy_capacity_mwh * 1e3,
        )

        # deactivate any objective functions
        for obj in blk.windBattery.component_objects(pyo.Objective):
            obj.deactivate()

        # initialize time index for this block
        b._time_idx = pyo.Param(initialize=0, mutable=True)

        active_blks = blk.windBattery.get_active_process_blocks()

        # create expression that references underlying power variables in multi-period rankine
        blk.HOUR = pyo.Set(initialize=range(horizon))
        blk.P_T = pyo.Expression(blk.HOUR)
        blk.tot_cost = pyo.Expression(blk.HOUR)
        blk.wind_waste_penalty = pyo.Param(default=1e3, mutable=True)
        blk.wind_waste = pyo.Expression(blk.HOUR)
        for (t, b) in enumerate(active_blks):
            blk.P_T[t] = (b.fs.splitter.grid_elec[0] + b.fs.battery.elec_out[0]) * 1e-3
            blk.wind_waste[t] = (b.fs.windpower.system_capacity * b.fs.windpower.capacity_factor[0] - b.fs.windpower.electricity[0]) * 1e-3
            blk.tot_cost[t] = b.fs.windpower.op_total_cost + b.fs.battery.var_cost + blk.wind_waste_penalty * blk.wind_waste[t]

        return

    def update_model(self, b, realized_soc, realized_energy_throughput):
        """Update variables using future wind capapcity the realized state-of-charge and enrgy throughput profiles.

        Args:
            b (block): the block that needs to be updated
            realized_soc (list): list of realized state of charge
            realized_energy_throughput (list): list of realized enrgy throughput
        """

        blk = b
        mp_wind_battery = blk.windBattery
        active_blks = mp_wind_battery.get_active_process_blocks()

        new_init_soc = round(realized_soc[-1], 2)
        active_blks[0].fs.battery.initial_state_of_charge.fix(new_init_soc)

        new_init_energy_throughput = round(realized_energy_throughput[-1], 2)
        active_blks[0].fs.battery.initial_energy_throughput.fix(
            new_init_energy_throughput
        )

        # shift the time -> update capacity_factor
        time_advance = min(len(realized_soc), 24)
        b._time_idx = pyo.value(b._time_idx) + time_advance

        new_capacity_factors = self._get_capacity_factors(b)
        update_wind_capacity_factor(mp_wind_battery, new_capacity_factors)

        return

    def _get_capacity_factors(self, b):
        """Fetch the future capacity factor.

        Args:
            b (block): the block that needs to be updated

        Returns:
            list: the capcity factors for the immediate future
        """

        horizon_len = len(b.windBattery.get_active_process_blocks())
        ans = self._wind_capacity_factors[
            pyo.value(b._time_idx) : pyo.value(b._time_idx) + horizon_len
        ]

        return ans

    @staticmethod
    def get_last_delivered_power(b, last_implemented_time_step):
        """Get last delivered power.

        Args:
            b (block): a multiperiod block
            last_implemented_time_step (int):  time index for the last implemented time period

        Returns:
            float: last delivered power
        """

        blk = b
        return pyo.value(blk.P_T[last_implemented_time_step])

    @staticmethod
    def get_implemented_profile(b, last_implemented_time_step):
        """Get implemented profiles, i.e., realized state-of-charge, energy throughput.

        Args:
            b (block): a multiperiod block
            last_implemented_time_step (int):  time index for the last implemented time period

        Returns:
            dict: dictionalry of implemented profiles.
        """

        blk = b
        mp_wind_battery = blk.windBattery
        active_blks = mp_wind_battery.get_active_process_blocks()

        realized_soc = deque(
            pyo.value(active_blks[t].fs.battery.state_of_charge[0])
            for t in range(last_implemented_time_step + 1)
        )

        realized_energy_throughput = deque(
            pyo.value(active_blks[t].fs.battery.energy_throughput[0])
            for t in range(last_implemented_time_step + 1)
        )

        return {
            "realized_soc": realized_soc,
            "realized_energy_throughput": realized_energy_throughput,
        }

    def record_results(self, b, date=None, hour=None, **kwargs):
        """Record the operations stats for the model, i.e., generator name, data, hour, horizon,
        total wind generation, total power output, wind power output, battery power output, charging power,
        state of charge, total costs.

        Args:
            b (block): a multiperiod block
            date (str, optional): current simulation date. Defaults to None.
            hour (int, optional): current simulation hour. Defaults to None.
        """
        blk = b
        mp_wind_battery = blk.windBattery
        active_blks = mp_wind_battery.get_active_process_blocks()

        df_list = []
        for t, process_blk in enumerate(active_blks):

            result_dict = {}

            result_dict["Generator"] = self.model_data.gen_name
            result_dict["Date"] = date
            result_dict["Hour"] = hour

            # simulation inputs
            result_dict["Horizon [hr]"] = int(t)

            # model vars
            result_dict["Total Wind Generation [MW]"] = float(
                round(pyo.value(process_blk.fs.windpower.electricity[0]) * 1e-3, 2)
            )
            result_dict["Total Power Output [MW]"] = float(
                round(pyo.value(blk.P_T[t]), 2)
            )
            result_dict["Wind Power Output [MW]"] = float(
                round(pyo.value(process_blk.fs.splitter.grid_elec[0] * 1e-3), 2)
            )
            result_dict["Wind Curtailment [MW]"] = float(
                round(pyo.value(blk.wind_waste[0]), 2)
            )
            result_dict["Battery Power Output [MW]"] = float(
                round(pyo.value(process_blk.fs.battery.elec_out[0] * 1e-3), 2)
            )
            result_dict["Wind Power to Battery [MW]"] = float(
                round(pyo.value(process_blk.fs.battery.elec_in[0] * 1e-3), 2)
            )
            result_dict["State of Charge [MWh]"] = float(
                round(pyo.value(process_blk.fs.battery.state_of_charge[0] * 1e-3), 2)
            )
            result_dict["Total Cost [$]"] = float(round(pyo.value(blk.tot_cost[t]), 2))

            for key in kwargs:
                result_dict[key] = kwargs[key]

            result_df = pd.DataFrame.from_dict(result_dict, orient="index")
            df_list.append(result_df.T)

        # append to result list
        self.result_list.append(pd.concat(df_list))

        return

    def write_results(self, path):
        """Write the saved results to a csv file.

        Args:
            path (str): the path to write the results.
        """

        pd.concat(self.result_list).to_csv(path, index=False)

    @property
    def power_output(self):
        return "P_T"

    @property
    def total_cost(self):
        return ("tot_cost", 1)
